#!/usr/bin/python

import sys
import types
import pyecm_wrapper
import Number, Factors

from gmpy import gcd, invert, mpz, next_prime, sqrt, root

indent = 4
verify = 1

def sgmftr(p,q):
    '''
    sigma( p, q ) = ( p ** ( q + 1 ) - 1 ) / ( p - 1 )
    '''

    if verbose:
        print "sgmftr: %s %s" % ( p, q )

    ans = factor_pqminus( p, q + 1 )
    if p > 2:
        ans = factors_divide( ans, factor_factor( p - 1 ) )

    if verbose:
        print "sgmftr: %s %s -> %s" % ( p, q, factors_as_string( ans ) )

    if verify:
        do_verify( ( p ** ( q + 1 ) - 1 ) / ( p - 1 ), ans )

    return ans

def do_verify( exp, factors ):
    
    got = factors_magnitude( factors )

    if exp != got:
        print "sgmftr error"
        print "    exp: %s" % exp
        print "    got: %s" % got
        print "    exp/got: %s" % ( exp / got )
        print "    got/exp: %s" % ( got / exp )
        print "    %s" % factors_as_string( factors )

def isMersenne( q ):
    f = Factors.Factors( )
    ans = f.isLucasLehmerPrime( q )

    quiet = True
    if not quiet:
        if ans == False and isinstance( ans, types.BooleanType ):
            status = "N" # not prime
            if verbose:
                print "%s: %s" % ( status, q )
        elif ans == 0:
            status = "P" # Prime, but not Mersenne prime
            if verbose:
                print "%s: %s" % ( status, q )

    if ans > 0:
        status = "M" # Mersenne prime
        if verbose:
            print "%s: %s -> %s" % ( status, q, ans )

    return ans

def factor_pqminus( p, q ):
    '''
    factor( p ** q - 1 )
    '''

    global verbose, level

    if verbose:
        level += indent
        print "%spqminus: %s %s" % ( " " * level, p, q )

    value = p ** q - 1
    if p == 2 and isMersenne( q ):
        ans = { }
        ans[ value ] = Number.Number( value, 1 )
        return ans

    if q == 1:
        ans = factor_factor( value )
    elif q == 2 and value < 4:
        ans = factor_factor( value )
    elif even( q ):
        n = q / 2
        if n == 1:
            # p^2-1 = (p-1)*(p+1)
            lfactors = factor_factor( p - 1 )
            rfactors = factor_factor( p + 1 )
        else:
            # p^2n-1 = (p^n-1)*(p^n+1)
            lfactors = factor_pqminus( p, n )
            rfactors = factor_pqplus( p, n )
#        ans = lfactors * rfactors
        ans = factors_multiply( lfactors, rfactors )
    else:
        # p^m*m-1 = (p^m-1)*(...)
        qfactors = factor_factor( q )
        keys = qfactors.keys( )
        if len( keys ) == 1:
            ans = factor_factor( value )
        else:
            ans = { }
            for exp in keys:
                if even( exp ):
                    continue
                tvalue = factor_pqminus( p, exp )
                for factor in tvalue.keys( ):
                    while value % factor == 0:
                        if not ans.has_key( factor ):
                            ans[ factor ] = Number.Number( factor, 1 )
                        else:
                            ans[ factor ] = Number.Number( factor, ans[ factor ].power( ) +  1 )
                        value = value / factor
            if value != 1:
                ans = factors_multiply( ans, factor_factor( value ) )

    if verbose:
        print "%spqminus: %s %s -> %s" % ( " " * level, p, q, factors_as_string( ans ) )
        level -= indent

    return ans

def factor_pqplus( p, q ):
    '''
    factor( p ** q + 1 )
    if q is odd:
        for each factor of q:
            merge factors of p**factor + 1 into result
    '''

    global verbose, level

    if verbose:
        level += indent
        print "%spqplus: %s %s" % ( " " * level, p, q )

    value = p ** q + 1
    if q == 1:
        ans = factor_factor( value )
    elif even( q ):
        ans = factor_factor( value )
    else:
        ans = { }
        qfactors = factor_factor( q )
        for exp in qfactors.keys( ):
            if even( exp ):
                continue
            tvalue = factor_factor( p ** exp + 1 )
            for factor in tvalue.keys( ):
                while value % factor == 0:
                    if not ans.has_key( factor ):
                        ans[ factor ] = Number.Number( factor, 1 )
                    else:
                        ans[ factor ] = Number.Number( factor, ans[ factor ].power( ) +  1 )
                    value = value / factor

        if value != 1:
            ans = factors_multiply( ans, factor_factor( value ) )

    if verbose:
        print "%spqplus: %s %s -> %s" % ( " " * level, p, q, factors_as_string( ans ) )
        level -= indent

    return ans

def factor_factor( v ):
    v = mpz( v )
    ans = pyecm_wrapper.factors_as_dict_of_numbers( v )
    if verbose:
        print "%sf_f: %s ->  %s" % (  " " * (level + indent), v, ans )
    return ans

def factor_factor_crude( v ):

    global verbose, level

    ans = { }
    n = 2
    while n * n < v:
        if v % n == 0:
            ans[ n ] = ans.get( n, 0 ) + 1
            v = v / n
        if v == 1:
            break
        n += 1
        
    if v != 1:
        ans[ v ] = ans.get( v, 0 ) + 1

    if verbose:
        level += indent
        print "%sfactor: %s -> %s" % ( " " * level, v, factors_as_string( ans ) )
        level -= indent

    return ans

def factors_multiply( lfactors, rfactors ):
    '''
    multiply two lists of factors
    '''
    lkeys = lfactors.keys( )
    rkeys = rfactors.keys( )
    if len( lkeys ) < len( rkeys ):
        ans = rfactors
        src = lfactors
    else:
        ans = lfactors
        src = rfactors

    for p in src.keys( ):
#        ans[ p ] = ans.get( p, 0 ) + src[ p ]
        if not ans.has_key( p ):
            q = src[ p ].power( )
        else:
            q = ans[ p ].power( ) + src[ p ].power( )
        ans[ p ] = Number.Number( p, q )

    if verbose:
        print "%smultiply: %s %s -> %s" % ( " " * level, lfactors, rfactors, factors_as_string( ans ) )

    return ans

def factors_divide( numerator, denominator ):
    '''
    divide two lists of factors
    '''

    if verbose:
        print "%sdivide: %s %s" % ( " " * level, numerator, denominator )

    ans = { }
    for key in numerator.keys( ):
        diff = numerator[ key ] - denominator.get( key, 0)
        if diff != 0:
            ans[ key ] = Number.Number( key, diff )

    if verbose:
        print "%sdivide: %s %s -> %s" % ( " " * level, numerator, denominator, factors_as_string( ans ) )

    return ans

def factors_as_string( factors ):
    ans = "("
    keys = factors.keys( )
    keys.sort( )
    for key in keys:
        ans = "%s %s^%s" % ( ans, key, factors[ key ].power( ) )
    ans = "%s )" % ( ans )
    return ans

def factors_magnitude( factors ):
    ans = 1
    keys = factors.keys( )
    for key in keys:
        ans = ans * key ** factors[ key ].power( )
    return ans

def even( n ):
    if n % 2 == 0:
        return True
    else:
        return False
    
def print_sgmftr( p, q, v ):
    s = pyecm_wrapper.time_now( )
    print "%s  " % s,
    print " %s  %s  %s" % ( p, q, factors_as_string( v ) )
    sys.stdout.flush( )

def main( ):

    for i in xrange( 1, len( sys.argv ), 2 ):
#        level = 0
        p = int( sys.argv[ i ] )
        q = int( sys.argv[ i + 1 ] )
        print "sgmftr( %s, %s ) -> " % ( p, q ),
        factors_as_string( sgmftr( p, q ) )

def do_mersenne( i, argc ):

    print argc, i
    if argc - i < 2:
        min = 2
        max = 200
    elif i < argc:
        min = int( sys.argv[ i + 0 ] )
        max = int( sys.argv[ i + 1 ] )

    for q in range( min, max ) :
        ans = isMersenne( q )
        if ans > 0:
            print q, ans

def do_test( i, argc ):

    if argc - i < 2:
        min = 2
        max = 200
    elif i < argc:
        min = int( sys.argv[ i + 0 ] )
        max = int( sys.argv[ i + 1 ] )
    
        two_68  = "07-08-20 09:56:12    2  68  ( 7^1 47^1 178481^1 10052678938039^1 )"
        two_137 = "07-10-02 18:08:03    2  137  ( 3^2 7^1 47^1 139^1 178481^1 2796203^1 168749965921^1 10052678938039^1 )"

        two_275_1 = "09-03-15 03:10:34    2  275  ( 3^2 5^1 7^1 13^1 47^1 139^1 277^1 1013^1 1657^1 30269^1 178481^1 2796203^1 168749965921^1 5415624023749^1 10052678938039^1                  70334392823809^1 )"
        two_275_2 = "09-03-18 13:03:22    2  275  ( 3^2 5^1 7^1 13^1 47^1       277^1 1013^1 1657^1 30269^1 178481^1 2796203^1                5415624023749^1                                   70334392823809^1 )"
        two_275_3 = "09-03-18 13:40:24    2  275  ( 3^2 5^1 7^1 13^1 47^1       277^1 1013^1 1657^1 30269^1 178481^1 2796203^1                5415624023749^1 10052678938039^1                  70334392823809^1 )"
        two_275_4 = "09-03-18 13:42:16    2  275  ( 3^2 5^1 7^1 13^1 47^1       277^1 1013^1 1657^1 30269^1 178481^1 2796203^1                5415624023749^1 10052678938039^1 23456245263019^1 70334392823809^1 )"

    p = 2
    for q in range( min, max ) :
#        if q == 275:
#            print two_275_4
#            print two_275_3
#            print two_275_2
#            print two_275_1
        print_sgmftr( p, q, sgmftr( p, q ) )

if __name__ == '__main__':
    global verbose, level

    level = 0
    verbose = 0

    argc = len(sys.argv)

    verbose = 0

    for i in range( 1, argc ):
        arg = sys.argv[ i ]
        if ( arg == "-v" ):
            verbose = 1
        elif ( arg == "t" ):
            do_test( i + 1, argc )
            break;
        elif ( arg == "m" ):
            do_mersenne( i + 1, argc )
            break;
